#(131663-104183) / 12

#18bits per WE, 64bits per line

import os, sys, time
import collections
import cProfile


dn = [  "I/O-0", "I/O-1", "I/O-2", "I/O-3", "I/O-4", "I/O-5", "I/O-6", "I/O-7", \
        "CLE", "ALE", "WE", "RE",]

NandCmd = collections.namedtuple("NandCmd", "name code addr_bytes code2" )
BusData = collections.namedtuple("BusData", "IO CLE ALE WE RE" )

CmdReset = NandCmd(name="Reset", code=0xFF, addr_bytes=0, code2=None )
CmdProgPage = NandCmd(name="ProgramPage", code=0x80, addr_bytes=5, code2=0x10 )
CmdReadId = NandCmd(name="ReadId", code=0x90, addr_bytes=1, code2=None )
CmdEraseBlock = NandCmd(name="EraseBlock", code=0x60, addr_bytes=3, code2=0xd0 )

all_cmds = {    CmdReset.code: CmdReset,
                CmdProgPage.code: CmdProgPage,
                CmdReadId.code:CmdReadId,
                CmdEraseBlock.code:CmdEraseBlock,
         }




def diff_bus_data(bd1, bd2):
    if bd1 is None or bd2 is None: return True
    return bd1.IO != bd2.IO or bd1.CLE != bd2.CLE \
            or bd1.ALE != bd2.ALE or bd1.WE != bd2.WE or bd1.RE != bd2.RE

class Bus:
    def __init__(self):
        self.signals = []
        self.last_signal_wr = None
        self.last_signal_rd = None

    def push(self, bdat):
        if diff_bus_data( bdat, self.last_signal_wr ):
            self.signals.append( bdat )
            self.last_signal_wr = bdat

    def pop(self):
        while len(self.signals) != 0:
            bdat = self.signals[0]
            self.signals = self.signals[1:]
            #return bdat
            # Detect rising edge of WE
            if self.last_signal_rd is not None:
                WE_rising = self.last_signal_rd.WE == 0 and bdat.WE == 1
                RE_rising = self.last_signal_rd.RE == 0 and bdat.RE == 1
                if WE_rising or RE_rising:
                    self.last_signal_rd = bdat
                    #print(" POP ", bdat, "%c" % bdat.IO)
                    return bdat
            self.last_signal_rd = bdat
        return None
        
bus = Bus()

def more_bits( bits_db ):
    minbits = len(bits_db[dn[0]])
    for v in bits_db.values():
        if len(v) < minbits:
            minbits = len(v)

    #print(minbits)
    # todo: only keep changes
    cle_bits = bits_db['CLE']
    ale_bits = bits_db['ALE']
    we_bits = bits_db['WE']
    re_bits = bits_db['RE']
    d0_bits = bits_db['I/O-0']
    d1_bits = bits_db['I/O-1']
    d2_bits = bits_db['I/O-2']
    d3_bits = bits_db['I/O-3']
    d4_bits = bits_db['I/O-4']
    d5_bits = bits_db['I/O-5']
    d6_bits = bits_db['I/O-6']
    d7_bits = bits_db['I/O-7']
    for i in range(minbits):
        ioval = 0
        if d0_bits[i] == '1': ioval += 0x01
        if d1_bits[i] == '1': ioval += 0x02
        if d2_bits[i] == '1': ioval += 0x04
        if d3_bits[i] == '1': ioval += 0x08
        if d4_bits[i] == '1': ioval += 0x10
        if d5_bits[i] == '1': ioval += 0x20
        if d6_bits[i] == '1': ioval += 0x40
        if d7_bits[i] == '1': ioval += 0x80

        cle = 1 if cle_bits[i] == '1' else 0
        ale = 1 if ale_bits[i] == '1' else 0
        we = 1 if we_bits[i] == '1' else 0
        re = 1 if re_bits[i] == '1' else 0

        bdat = BusData( IO=ioval, CLE=cle, ALE=ale, WE=we, RE=re)
        """
        bits = [ bits_db[k][i] for k in dn[:8] ]
        val = 0
        for b in reversed(bits):
            val <<= 1
            if b == '1':
                val |= 1
        #print("IO: ", val)
        
        bdat = BusData( IO=val, CLE=int(bits_db['CLE'][i]), ALE=int(bits_db['ALE'][i]),
                        WE=int(bits_db['WE'][i]), RE=int(bits_db['RE'][i]) )
        """

        bus.push( bdat )

    # truncate bits that were used
    for k, v in bits_db.items():
        bits_db[ k ] = v[minbits:]
    
    bus_consumer.next()


def more_bus_data():

    while True:
        bdat = bus.pop()
        if bdat is None:
            yield 0
            continue

        #print(bdat, "%c" % bdat.IO)
        if bdat.CLE == 1 and bdat.IO in all_cmds:
            cmd = all_cmds[ bdat.IO ]
            print("CMD start", cmd.name)
            cnt = cmd.addr_bytes
            col_addr = row_addr = 0
            code2 = None
            while cnt != 0:
                addr = bus.pop()
                if addr is None: yield 0
                elif addr.ALE==0: raise Exception("Unexpected non addr")
                else:
                    if cnt > 3: col_addr |= addr.IO << (8*(cmd.addr_bytes-cnt))
                    else:       row_addr |= addr.IO << (8*(3-cnt))
                    cnt -= 1
            #print("addr col %d row %d", col_addr, row_addr)
            writeData = ""
            if cmd.code2 is not None:
                while True:
                    bdat = bus.pop()
                    if bdat is None: yield 0
                    elif bdat.ALE==1: raise Exception("Unexpected addr")
                    #elif bdat.WE==1 and bdat.RE==1: raise Exception("Unexpected RE+WE")
                    #elif bdat.RE==1: readData += "%c"%bdat.IO
                    elif bdat.CLE==1 and bdat.IO == cmd.code2:
                        code2 = bdat
                        break
                    elif bdat.WE==1: writeData += "%c"%bdat.IO
            print("CMD complete ", cmd, "addr: c=%x,r=%x"%(col_addr,row_addr), "code2:", code2,
                    "write:", len(writeData)) #, writeData)
            writeFirmwareBlock(col_addr, row_addr, writeData)
        elif bdat.CLE == 1 and bdat.IO not in all_cmds:
            print("UNKNOWN", bdat, "%c" % bdat.IO)
            sys.stderr.write("UNKNOWN", bdat, "%c" % bdat.IO, "\n")
        elif bdat.RE == 1:
            sys.stdout.write( "%c"%bdat.IO )
        else:
            pass #print(bdat, "%c" % bdat.IO)

total_block = 0x1200
cur_block = 0
t0 = time.time()
frmf = open("firmware3.bin", "wb+")
def writeFirmwareBlock(col_addr, row_addr, writeData):
    # written in row first the col (like a pixel image)
    block_addr = col_addr * (256*256*256) + row_addr
    # Blocks seems to be 8Kb large
    lin_addr = block_addr * 8192
    frmf.seek( lin_addr, 0 ) # from beginning
    frmf.write(writeData)
    #frmf.flush()

    global cur_block, total_block, t0
    cur_block += 1
    pc = 100 * cur_block / total_block
    eta = int(((time.time() - t0) / cur_block) * (total_block-cur_block))
    h = eta//3600
    m = (eta - h*3600) // 60
    sys.stderr.write("cur_block: %d/%d (%.1f%%) ETA %02dh%02dm\n" % (cur_block, total_block, pc, h, m))


bus_consumer = more_bus_data( )

l = sys.stdin.readline()
l = sys.stdin.readline()
lrd = 2

db = { n:[] for n in dn }
done = False
while not done:
     
    for i in range(12):
        l = sys.stdin.readline().strip()
        if l is None or len(l) == 0:
            done = True
            break
        lrd += 1
        #print(l)
        nm, bits = l.split(':')
        bits = str(bits.strip().replace(' ', ''))
        db[nm] += bits 

    more_bits( db )
    
    
