
import struct,sys

txt = sys.stdin.read()

n = len(txt) / 8
s = ""

s += "%u" % n
for i in range(n):
    op, jt, jf, k = struct.unpack("<HBBI", txt[i*8:i*8+8]) 
    s += ", %u %u %u %u" % (op, jt, jf, k)
#print(s)
    
"""

845   +----------------+--------+--------------------+
846   |   4 bits       |  1 bit |   3 bits           |
847   | operation code | source | instruction class  |
848   +----------------+--------+--------------------+
849   (MSB)                                      (LSB)
850 
851 Three LSB bits store instruction class which is one of:
852 
853   Classic BPF classes:    eBPF classes:
854 
855   BPF_LD    0x00          BPF_LD    0x00
856   BPF_LDX   0x01          BPF_LDX   0x01
857   BPF_ST    0x02          BPF_ST    0x02
858   BPF_STX   0x03          BPF_STX   0x03
859   BPF_ALU   0x04          BPF_ALU   0x04
860   BPF_JMP   0x05          BPF_JMP   0x05
861   BPF_RET   0x06          [ class 6 unused, for future if needed ]
862   BPF_MISC  0x07          BPF_ALU64 0x07
863 
864 When BPF_CLASS(code) == BPF_ALU or BPF_JMP, 4th bit encodes source operand ...
865 
866   BPF_K     0x00
867   BPF_X     0x08
868 
869  * in classic BPF, this means:
870 
871   BPF_SRC(code) == BPF_X - use register X as source operand
872   BPF_SRC(code) == BPF_K - use 32-bit immediate as source operand
873 
874  * in eBPF, this means:
875 
876   BPF_SRC(code) == BPF_X - use 'src_reg' register as source operand
877   BPF_SRC(code) == BPF_K - use 32-bit immediate as source operand
878 
879 ... and four MSB bits store operation code.
880 
881 If BPF_CLASS(code) == BPF_ALU or BPF_ALU64 [ in eBPF ], BPF_OP(code) is one of:
882 
883   BPF_ADD   0x00
884   BPF_SUB   0x10
885   BPF_MUL   0x20
886   BPF_DIV   0x30
887   BPF_OR    0x40
888   BPF_AND   0x50
889   BPF_LSH   0x60
890   BPF_RSH   0x70
891   BPF_NEG   0x80
892   BPF_MOD   0x90
893   BPF_XOR   0xa0
894   BPF_MOV   0xb0  /* eBPF only: mov reg to reg */
895   BPF_ARSH  0xc0  /* eBPF only: sign extending shift right */
896   BPF_END   0xd0  /* eBPF only: endianness conversion */
897 
898 If BPF_CLASS(code) == BPF_JMP, BPF_OP(code) is one of:
899 
900   BPF_JA    0x00
901   BPF_JEQ   0x10
902   BPF_JGT   0x20
903   BPF_JGE   0x30
904   BPF_JSET  0x40
905   BPF_JNE   0x50  /* eBPF only: jump != */
906   BPF_JSGT  0x60  /* eBPF only: signed '>' */
907   BPF_JSGE  0x70  /* eBPF only: signed '>=' */
908   BPF_CALL  0x80  /* eBPF only: function call */
909   BPF_EXIT  0x90  /* eBPF only: function return */

931 For load and store instructions the 8-bit 'code' field is divided as:
932 
933   +--------+--------+-------------------+
934   | 3 bits | 2 bits |   3 bits          |
935   |  mode  |  size  | instruction class |
936   +--------+--------+-------------------+
937   (MSB)                             (LSB)
938 
939 Size modifier is one of ...
940 
941   BPF_W   0x00    /* word */
942   BPF_H   0x08    /* half word */
943   BPF_B   0x10    /* byte */
944   BPF_DW  0x18    /* eBPF only, double word */
945 
946 ... which encodes size of load/store operation:
947 
948  B  - 1 byte
949  H  - 2 byte
950  W  - 4 byte
951  DW - 8 byte (eBPF only)
952 
953 Mode modifier is one of:
954 
955   BPF_IMM  0x00  /* used for 32-bit mov in classic BPF and 64-bit in eBPF */
956   BPF_ABS  0x20
957   BPF_IND  0x40
958   BPF_MEM  0x60
959   BPF_LEN  0x80  /* classic BPF only, reserved in eBPF */
960   BPF_MSH  0xa0  /* classic BPF only, reserved in eBPF */
961   BPF_XADD 0xc0  /* eBPF only, exclusive add */
962 
963 eBPF has two non-generic instructions: (BPF_ABS | <size> | BPF_LD) and
964 (BPF_IND | <size> | BPF_LD) which are used to access packet data.

"""

cls_name = {
    0x00: 'BPF_LD',
    0x01: 'BPF_LDX',
    0x02: 'BPF_ST',
    0x03: 'BPF_STX',
    0x04: 'BPF_ALU',
    0x05: 'BPF_JMP',
    #0x06: 'BPF_RET',
    0x07: 'BPF_ALU64',
}
op_name_alu = {
    0x00: 'BPF_ADD',
    0x10: 'BPF_SUB',
    0x20: 'BPF_MUL',
    0x30: 'BPF_DIV',
    0x40: 'BPF_OR',
    0x50: 'BPF_AND',
    0x60: 'BPF_LSH',
    0x70: 'BPF_RSH',
    0x80: 'BPF_NEG',
    0x90: 'BPF_MOD',
    0xa0: 'BPF_XOR',
    0xb0: 'BPF_MOV',
    0xc0: 'BPF_ARSH',
    0xd0: 'BPF_END',
}
c_op_name_alu = {
    0x00: ' += ',
    0x10: ' -= ',
    0x20: ' *= ',
    0x30: ' /= ',
    0x40: ' |= ',
    0x50: ' &= ',
    0x60: ' <<= ',
    0x70: ' >>= ',
    0x80: ' =- ',
    0x90: ' %= ',
    0xa0: ' ^= ',
    0xb0: ' = ',
    0xc0: '???BPF_ARSH',
    0xd0: '???BPF_END',
}

op_name_jmp = {
    0x00: 'BPF_JA',
    0x10: 'BPF_JEQ',
    0x20: 'BPF_JGT',
    0x30: 'BPF_JGE',
    0x40: 'BPF_JSET',
    0x50: 'BPF_JNE',
    0x60: 'BPF_JSGT',
    0x70: 'BPF_JSGE',
    0x80: 'BPF_CALL',
    0x90: 'BPF_EXIT',
    #0xb0: 'BPF_????????????',
}
c_op_name_jmp = {
    0x00: 'goto {2}; // unused {0}, {1}',
    0x10: 'if( {0} == {1} ) goto {2};',
    0x20: 'if( {0} > {1} ) goto {2};',
    0x30: 'if( {0} >= {1} ) goto {2};',
    0x40: '???BPF_JSET',
    0x50: 'if( {0} != {1} ) goto {2};',
    0x60: '???BPF_JSGT',
    0x70: '???BPF_JSGE',
    0x80: '???BPF_CALL',    # handled as a special case
    0x90: 'return R0; // unused {0}, {1}, {2}',
    #0xb0: 'BPF_????????????',
}

src_name = {
    0x00: 'BPF_K',  # use immediate val
    0x08: 'BPF_X',  # use in src_reg
}

size_name = {
    0x00: 'word',
    0x08: 'halfword',
    0x10: 'byte',
    0x18: 'dblword',
}
c_type_name = {
    0x00: 'uint32_t',
    0x08: 'uint16_t',
    0x10: 'uint8_t',
    0x18: 'uint64_t',
}

mode_name = {
    0x00: 'BPF_IMM',
    0x20: 'BPF_ABS',
    0x40: 'BPF_IND',
    0x60: 'BPF_MEM',
    0x80: 'BPF_LEN!!!!!',   # only classic BPF
    0xa0: 'BPF_MSH!!!!!',   # only classic BPF
    0xc0: 'BPF_XADD',
}

def decode(instru):
    registers = ord(instru[1])
    dst_reg = registers&0xF
    src_reg = (registers>>4)&0xF
    off = instru[2:4]
    #jtgt, = struct.unpack("<H", off)
    offset, = struct.unpack("<h", off)
    imm = instru[4:8]
    imval, = struct.unpack("<i", imm)
    opcode = ord(instru[0])
    cls = opcode & 0x7
    return cls, opcode, dst_reg, src_reg, offset, imval

def bpf_patcher(prog, table_fd):
    new=""
    endian="big"
    if struct.pack("=I",1) == "\x01\x00\x00\x00":
        endian="little"

    jmp_targets = set()
    # I know those 2 are jump backwards
    jmp_targets.add( 110 )
    jmp_targets.add( 330 )
    i = 0
    #for i in range(0,len(prog),8):
    while i < len(prog):
        extra = False
        cur = i/8
        nxt = cur + 1
        cls, opcode, dst_reg, src_reg, offset, imval = decode( prog[i:i+8] )
        clsnam = cls_name[cls]
        s = "%-12s" % clsnam
        c = ""
        if cls <= 3:    # LD and ST classes
            mode = opcode & 0xe0
            size = opcode & 0x18
            modenam = mode_name[mode]
            sizenam = size_name[size]
            s += " %s mode:%s" % (sizenam, modenam)
            s += ", dst:R%d" % (dst_reg)
            if modenam == 'BPF_IMM':    s += " IMM 0x%0x (%d)" % (imval, imval)
            elif modenam == 'BPF_ABS':  s += " R6->data + IMM 0x%0x (%d)" % (imval, imval)
            elif modenam == 'BPF_IND':  s += " R6->data + src:R%d + IMM 0x%0x (%d)" % (src_reg, imval, imval)
            else:                       s += " src:R%d IMM 0x%0x" % (src_reg, imval)
            if opcode == 0x18 and src_reg != 0:
                s += "        ** TABLEID"

            cdst = "R%d" % (dst_reg)
            csrc = "R%d" % (src_reg)
            ctyp = c_type_name[size]
            if clsnam == 'BPF_LD':
                if opcode == 0x18 and src_reg != 0:
                    c += "%s = 12345;" % (cdst)
                    i += 8  # special insn_ load tableid takes 2 insn, skip the next one
                    extra = True
                elif modenam == 'BPF_IMM' and sizenam == 'dblword':
                    i += 8  # special insn_ load doubleword takes 2 insn, skip the next one
                    cls2, opcode2, _, _, _, imval2 = decode( prog[i:i+8] )
                    # check this 2nd insn is BPF_LD sizes:=word mode:BPF_IMM
                    assert(cls2 == 0x00 and opcode2 == 0x00)
                    #if imval < 0:
                    #    imval = (~imval) + 1  # convert this imm to unsigned since it is the lower 32bits
                    imval |= imval2 << 32;
                    c += "%s = %ld;" % (cdst, imval)
                    #print( "extra insn:", prog[i:i+8] )
                    extra = True
                elif modenam == 'BPF_IMM':  c += "%s = %d;" % (cdst, imval)
                elif modenam == 'BPF_ABS':  c += "%s = load_be_%s(R6 + %d);" % (cdst, ctyp, imval)
                elif modenam == 'BPF_IND':  c += "%s = load_be_%s(R6 + %s + %d);" % (cdst, ctyp, csrc, imval)
                elif modenam == 'BPF_MEM':  assert(False); c += "%s = *(%s *)(R6 + %s + %d);" % (cdst, ctyp, csrc, imval)
            elif clsnam == 'BPF_LDX':
                assert( modenam == 'BPF_MEM' )
                c += "%s = *(%s *)(%s + %d);" % (cdst, ctyp, csrc, offset)
            else:
                assert( clsnam == 'BPF_STX' and modenam == 'BPF_MEM' )
                c += "*(%s *)(%s + %d) = %s;" % (ctyp, cdst, offset, csrc)

        elif cls == 5:  # JMP insn
            op = opcode & 0xf0
            src = opcode & 0x8
            opnam = op_name_jmp[op]
            srcnam = src_name[src]
            s += " %s" % (opnam)
            if opnam == 'BPF_CALL':
                if srcnam == 'BPF_X':   s += " src:R%d" % (src_reg)
                else:                   s += " IMM 0x%0x (%d)" % (imval, imval)

                c += "R0 = func_%d(R1, R2, R3, R4);" % (imval)
            else:
                s += ", R%d" % (dst_reg)
                if srcnam == 'BPF_X':   s += " src:R%d" % (src_reg)
                else:                   s += " IMM 0x%0x (%d)" % (imval, imval)
                s += "     ------> I%04d (%+d)" % (nxt+offset, offset)
                if offset > 0:
                    jmp_targets.add( nxt+offset )

                cdst = "R%d" % (dst_reg)
                if srcnam == 'BPF_X':   csrc = "R%d" % (src_reg)
                else:                   csrc = "%d" % (imval)
                cjtgt = "I%04d" % (nxt+offset)

                c += c_op_name_jmp[op].format( cdst, csrc, cjtgt )
        else:
            assert( clsnam == 'BPF_ALU' or clsnam == 'BPF_ALU64' )
            op = opcode & 0xf0
            src = opcode & 0x8
            opnam = op_name_alu[op]
            srcnam = src_name[src]
            s += " %s" % (opnam)
            s += ", dst:R%d" % (dst_reg)
            if srcnam == 'BPF_X':   s += " src:R%d" % (src_reg)
            else:                   s += " IMM 0x%0x (%d)" % (imval, imval)

            cdst = "R%d" % (dst_reg)
            if srcnam == 'BPF_X':   csrc = "R%d" % (src_reg)
            else:                   csrc = "%d" % (imval)

            cop = c_op_name_alu[op]
            c = cdst + cop + csrc + ";"
            
        if cur in jmp_targets:  sp = "I%04u:   " % cur
        else:                   sp = "     .   "
            
        if cur in jmp_targets:  csp = "I%04u:   " % cur
        else:                   csp = "         "
            
        if cur in jmp_targets:  jmp_targets.remove(cur)

        if len(c) == 0:
            c = "?????????" # put an incompatible C statement to catch non-translated insn
        #print( "// " + sp + s )
        #print( csp + c )
        print( "%-55s // %s" % (csp + c, s) )
        if extra:
            print( "%-55s // %s" % ("", "long insn") )
        i += 8

    assert( len(jmp_targets) == 0 )
    for jtgt in jmp_targets:
        csp = "I%04u:   " % jtgt
        print( csp )
    print( "exit(-1); // undefined jump target" )

    return new

bpf_patcher(txt, 0)

