
// gcc -m32 -O0 decrypt.c  -o decrypt

#include <stdio.h>
#include <assert.h>
#include <inttypes.h>
#include <time.h>
#include <string.h>

typedef uint8_t u8;
typedef uint32_t u32;

typedef struct bitset128_ {
    u8 m[16];
} BitSet128;

typedef struct _bitgroup {
    u8 numPasswdBits;
    u8 passwdBitPos[8];       // max 8 bits per group
    BitSet128 outMask;
} BitGroup;

typedef struct _ctx {
    u8 passwd[16], in[16], tgt[16];
    u8 block_num;

    u8 numBitGroups;        // max 256
    BitGroup bitGroup[256];
} Ctx;


static void BSClear(BitSet128 *bm) {
    int i=0;
    for(i=0; i<16; i++) bm->m[i] = 0;
}
static void BSAssign(BitSet128 *bm, const u8 *v) {
    int i=0;
    for(i=0; i<16; i++) bm->m[i] = v[i];
}
static void BSSetBit(BitSet128 *bm, const int i) {
    bm->m[i/8] |= 1<<(i%8);
}
static void BSClearBit(BitSet128 *bm, const int i) {
    u8 b = bm->m[i/8];
    b &= ~(1<<(i%8));
    bm->m[i/8] = b;
}
static int BSIsBitSet(BitSet128 *bm, const int i) {
    return (bm->m[i/8] & (1<<(i%8))) != 0;
}
static int BSGetBit(BitSet128 *bm, const int i) {
    return bm->m[i/8] >> (i%8);
}


static void pp_array(const char *head, u8 *arr, int len) {
    int i;
    if(head) printf("%s", head);
    for(i=0; i<len; i++)
        printf("%02x ", arr[i]);
    printf("\n");
}
static void pp_array_compact(const char *head, u8 *arr, int len) {
    int i;
    if(head) printf("%s", head);
    for(i=0; i<len; i++)
        printf("%02x", arr[i]);
    printf("\n");
}

static int decrypt( const u8 *passwd, u32 block_num, const u8 *inptr, u8* outptr) {
    int i;
    u32 a0, a1, a2, a3, a4, a5, a6, a7;
    u32 s0, s1, s2, s3, s4, s5, s6, s7;
    u32 t0, t1, t2, t3, t4, t5, t6, t7;

    u8 sp[512];

    s0 = (u32)(inptr);
    s1 = block_num;

    a3 = *(u32*)(passwd+4);
    t5 = *(u32*)(passwd+12);
    t4 = *(u32*)(passwd+0);
    t2 = *(u32*)(passwd+8);

    t0 = 0x52555655;
    s2 = 0xadaaa9aa;

    #include "decrypt_gen_1.c"

    for(i=0; i<16; i++) {

        outptr[ i ] = inptr[ i ] ^ sp[ 64 + i ];
    }
}

static void randomize( u8 *ptr ) {
    int i;
    for(i=0; i<16; i++) {
        ptr[i] = (u8)(random() % 0xff);
    }
}

static int detect(Ctx *ctx) {
    int i, j, exp;
    int maxexp = 1000;

    u8 in[16], passwd[16]; //, out[16], out2[16];
    //u8 in2OutChangeMask[16][16];
    u8 numinbitperoutbyte[16];
    memset(numinbitperoutbyte, 0, 16);
    //memset(in2OutChangeMask, 0, 16*16);
    BitSet128 out, out2;
    BitSet128 in2OutChangeMask[128];
    for(j=0; j<128; j++)
        BSClear( &in2OutChangeMask[j] );

    // Iterate over all bits of the password
    for(i=0; i<16*8; i++) {
        int cnt = 0;
        BitSet128 outchangemask;
        BSClear( &outchangemask );
        //u8 outchangemask[16];
        //memset(outchangemask, 0, 16);
        for(exp=0; exp<maxexp; exp++) {
            randomize( passwd );
            randomize( in );

            decrypt( passwd, ctx->block_num, in, out.m );
            
            u8 mask = 1 << (i%8);
            u8 b = passwd[i/8];
            if( (b & mask) == 0 )
                b |= mask;
            else 
                b &= ~mask; 
            passwd[i/8] = b;

            decrypt( passwd, ctx->block_num, in, out2.m );

            // Iterate over all bits of the output
            for(j=0; j<16*8; j++) {
                //u8 mask = 1 << (j%8);
                //if( (out[j/8] & mask) != (out2[j/8] & mask) ) {
                if( BSGetBit( &out, j ) != BSGetBit( &out2, j ) ) {
                    BSSetBit( &outchangemask, j );
                    //outchangemask[j/8] |= mask;
                    cnt++;
                }
            }
        }

        // count number of bits influencing this out byte
        //for(j=0; j<16; j++) {
        //    if( outchangemask[j] != 0)
        //        numinbitperoutbyte[j] ++;
        //}

        for(j=0; j<128; j++) {
            if( BSIsBitSet( &outchangemask, j ) )
                BSSetBit( &in2OutChangeMask[i], j );
        }
        //for(j=0; j<16; j++)
        //    in2OutChangeMask[i/8][j] |= outchangemask[j];

        printf("%04x: %6u ", i, cnt);
        pp_array(NULL, outchangemask.m, 16);
    }

    // Pretty print the results
    printf("----: %6u ", 0);
    pp_array(NULL, numinbitperoutbyte, 16);



    // Reverse the findings: for each output bit, find the set (mask) of
    // inout bit that can affect it
    BitSet128 Out2InChangeMask[128];
    for(j=0; j<16*8; j++) { // Iterate over all bits of the output
        BSClear( &Out2InChangeMask[j] );
        for(i=0; i<16*8; i++) {
            //if( (in2OutChangeMask[i/8][j/8] & (1<<(j%8))) != 0 ) 
            if( BSIsBitSet( &in2OutChangeMask[i], j ) ) 
                BSSetBit( &Out2InChangeMask[j], i );
                //Out2InChangeMask[j][i/8] |= (1 << (i%8)); 
        }
    }

    // Build bitGroups, simplified:
    // for each output byte, build the set of inbits that can affect it and
    // has not been used in a previous BitGroup
    //u8 usedPasswdBits[16];
    //memset(usedPasswdBits, 0, 16);
    BitSet128 usedPasswdBits;
    BSClear( &usedPasswdBits );

    int b, bgnum=0;
    for(j=15; j>=0; j--) { // Iterate over all bytes of the output
        BitGroup *bg = &(ctx->bitGroup[bgnum]);
        bg->numPasswdBits = 0;
        memset(bg->passwdBitPos, 0, 8);
        BSClear( &bg->outMask );
        //memset(bg->outMask, 0, 16);

        // Simpified: granularity is one out byte
        for(b=0; b<8; b++)
            BSSetBit( &bg->outMask, j*8+b );
        //bg->outMask[j] = 0xff;

        for(b=0; b<8; b++) {    // For each bits of the output byte
            for(i=0; i<16*8; i++) {
                // Check that passwd bit has not yet been used in a prev group
                //if( (usedPasswdBits[i/8] & (1<<(i%8))) != 0)
                if( BSIsBitSet( &usedPasswdBits, i ) )
                    continue;
                //if( (Out2InChangeMask[(j*8) + b][i/8] & (1<<(i%8))) != 0) {
                if( BSIsBitSet( &Out2InChangeMask[(j*8) + b], i ) ) {
                    assert(bg->numPasswdBits < 8 );
                    bg->passwdBitPos[bg->numPasswdBits] = i;
                    bg->numPasswdBits++;

                    //usedPasswdBits[i/8] |= (1<<(i%8));
                    BSSetBit( &usedPasswdBits, i );
                }
            }
        }

        bgnum++;
    }
    ctx->numBitGroups = bgnum;
    assert(bgnum == 16 );


    //for(i=0; i<16; i++) assert(usedPasswdBits[i] == 0xff);
    for(i=0; i<128; i++) assert( BSIsBitSet( &usedPasswdBits, i ) );

    for(i=0; i<ctx->numBitGroups; i++) {
        printf("%04u: ", i);
        BitGroup *bg = &ctx->bitGroup[i];
        for(j=0; j<bg->numPasswdBits; j++) {
            printf("%02x ", bg->passwdBitPos[j]);
        }
        printf("\n");
    }

    for(i=0; i<16; i++) {
        printf("%4u:  ", i);
        pp_array(NULL, in2OutChangeMask[i].m, 16);
    }
}

static int searchBitGroup(Ctx *ctx, int bitGroupNum) {
    u32 i, j, n;
    u8 out[ 16 ];

    BitGroup* bg = &( ctx->bitGroup[ bitGroupNum ] );
 
    for( i=0; i<(1<<bg->numPasswdBits); i++) {

        // Set bits in password
        for(n=0; n<bg->numPasswdBits; n++) {
            u8 bp = bg->passwdBitPos[n];
            ctx->passwd[bp/8] &= ~ (1 << (bp%8));  // clear bit
            if( (i&(1<<n)) != 0 )
                ctx->passwd[bp/8] |= (1 << (bp%8));  // set bit
        }

        // decrypt
        decrypt( ctx->passwd, ctx->block_num, ctx->in, out );

        // Check that the target bits controled by this group have the target value,
        for(j=0; j<16; j++) {
            //if( (out[j] & bg->outMask[j]) != (ctx->tgt[j] & bg->outMask[j]) ) {
            if( (out[j] & bg->outMask.m[j]) != (ctx->tgt[j] & bg->outMask.m[j]) ) {
                break;
            }
        }

        // Move on to next bitgroup if target is right continue search otherwise
        if( j == 16 ) {
            if(bitGroupNum+1 >= ctx->numBitGroups) {
                pp_array_compact("Candidate: ", ctx->passwd, 16);
            }
            else {
                searchBitGroup( ctx, bitGroupNum+1 );
            }
        }
    }

    return -1;
}

static int searchPasswd() {
    int i;
    Ctx ctx;
/*
    // first line of password_is_00112233445566778899AABBCCDDEEFF.txt.encrypted
    int block_num = 0;
    u8 in[16]  = "\x36\x36\x4c\x47\xf9\xd6\x48\x39\x84\xd7\xd2\x53\xdb\x47\xb2\x21";
    u8 ver[16] = "\x44\x65\x63\x72\x79\x70\x74\x65\x64\x20\x62\x6c\x6f\x63\x6b\x0a";
    u8 xor[16] = "\x35\xf8\x2d\xea\x31\x13\xd2\xec\xf3\xed\x35\x54\xc4\xa1\x99\xf6";

    // first line of secret.lzma.encrypted
    int block_num = 0;
    u8 in[16]  = "\x88\xdb\x9a\x97\x33\x65\x5c\xbf\x06\x3b\x59\x02\x44\xf0\x75\xdb";
    u8 ver[16] = "\x5d\x00\x00\x40\x00\xff\xff\xff\xff\xff\xff\xff\xff\x00\x3a\x1a";
    u8 xor[16] = "\x39\x96\x2d\x55\xd1\x38\xef\xef\xc6\xe4\xa1\x4d\x5c\xd7\x35\x2e";

    // last line of password_is_00112233445566778899AABBCCDDEEFF.txt.encrypted
    int block_num = 16;
    u8 in[16]  = "\xaa\x13\xa7\x95\x16\xb9\xa4\x47\x86\xb0\x58\xec\xc4\x60\xa9\x6f";
    u8 ver[16] = "\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10";
    u8 xor[16] = "\xa5\x98\x3d\x8a\x2e\xe3\xc9\x9a\x97\x47\x0d\xd2\x87\x19\xda\x3e";

    // last line of secret.lzma.encrypted
    int block_num = 2221;
    u8 in[16]  = "\xe6\x53\xff\x9e\x18\xc8\x8d\x77\x47\x9a\x91\x79\xd9\x84\x04\xc4";
    u8 ver[16] = "\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10";
    u8 xor[16] = "\x17\xf7\x5c\xe6\xd6\x4c\x28\x3b\x0e\x3a\xbe\xd1\x2e\x80\x62\x4a";

    // first line of secret.lzma.encrypted
    int block_num = 0;
    u8 in[16]  = "\x88\xdb\x9a\x97\x33\x65\x5c\xbf\x06\x3b\x59\x02\x44\xf0\x75\xdb";
    u8 ver[16] = "\x5d\x00\x00\x20\x00\xff\xff\xff\xff\xff\xff\xff\xff\x00\x3a\x1a";
    u8 xor[16] = "\x39\x96\x2d\x55\xd1\x38\xef\xef\xc6\xe4\xa1\x4d\x5c\xd7\x35\x2e";
*/
    // first line of password_is_00112233445566778899AABBCCDDEEFF.txt.encrypted
    int block_num = 0;
    u8 in[16]  = "\x88\xdb\x9a\x97\x33\x65\x5c\xbf\x06\x3b\x59\x02\x44\xf0\x75\xdb";
    u8 ver[16] = "\x5d\x00\x00\x80\x00\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00";
    u8 xor[16] = "\x39\x96\x2d\x55\xd1\x38\xef\xef\xc6\xe4\xa1\x4d\x5c\xd7\x35\x2e";

    for(i=0; i<16; i++) ver[i] ^= xor[15-i];

    ctx.block_num = block_num;
    detect( &ctx );

    memset(ctx.passwd, 0, 16);
    memcpy(ctx.in, in, 16);
    memcpy(ctx.tgt, ver, 16);

    // Trick to iterate throuh the bits that affect the last 2 bytes of the result, 
    // since those are not really guessed. Setting hte outMask to 0 means that the
    // matching will be ok for all bits in these groups hence giving many candidates 
    BSClear( &ctx.bitGroup[0].outMask );
    BSClear( &ctx.bitGroup[1].outMask );

    searchBitGroup(&ctx, 0);
}

int main(int argc, char *argv[]) {
    int i;

#if 1
    time_t t = time( NULL );
    srandom( t );

    searchPasswd( );
    return 0;
#endif

    u8 out[17];

#if 0
    // block 0
    int block_num = 0;
    u8 passwd[16] = "\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb\xcc\xdd\xee\xff";
    u8 in[16]  = "\x36\x36\x4c\x47\xf9\xd6\x48\x39\x84\xd7\xd2\x53\xdb\x47\xb2\x21";
    //u8 ver[16] = "\xb2\xfc\xc2\xb6\x2d\x45\x99\x96\x88\xf2\x71\x5d\x85\x4e\x93\x3f";
    u8 ver[16] = "\x44\x65\x63\x72\x79\x70\x74\x65\x64\x20\x62\x6c\x6f\x63\x6b\x0a";
    u8 xor[16] = "\x35\xf8\x2d\xea\x31\x13\xd2\xec\xf3\xed\x35\x54\xc4\xa1\x99\xf6";
#endif

    // block 1
#if 0
    int block_num = 1;
    u8 passwd[16] = "\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb\xcc\xdd\xee\xff";
    u8 in[16]  = "\x8d\x61\x63\x02\xf8\xc9\x64\x17\xb5\xd4\x72\x5f\xec\x40\x09\x13";
    //u8 ver[16] = "\x65\xd7\x24\xa9\x2a\xa2\xa3\xe1\x5d\x68\xb4\x95\x28\x2f\x5d\x3c";
    u8 ver[16] = "\x44\x65\x63\x72\x79\x70\x74\x65\x64\x20\x62\x6c\x6f\x63\x6b\x0a";
    u8 xor[16] = "\x36\x36\x4c\x47\xf9\xd6\x48\x39\x84\xd7\xd2\x53\xdb\x47\xb2\x21";
#endif
    
#if 0
    // block 16
    int block_num = 16;
    u8 passwd[16] = "\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb\xcc\xdd\xee\xff";
    u8 in[16]  = "\xaa\x13\xa7\x95\x16\xb9\xa4\x47\x86\xb0\x58\xec\xc4\x60\xa9\x6f";
    u8 ver[16] = "\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10\x10";
    u8 xor[16] = "\xa5\x98\x3d\x8a\x2e\xe3\xc9\x9a\x97\x47\x0d\xd2\x87\x19\xda\x3e";
#endif

    int block_num = 0;
    u8 passwd[16] = "\x51\x29\xcd\x9f\xd2\xec\x7b\xd9\x24\x4e\xce\x53\x8a\xce\xcb\xdf";
    u8 in[16]  = "\x88\xdb\x9a\x97\x33\x65\x5c\xbf\x06\x3b\x59\x02\x44\xf0\x75\xdb";
    u8 ver[16] = "\x5d\x00\x00\x20\x00\xff\xff\xff\xff\xff\xff\xff\xff\x00\x3a\x1a";
    u8 xor[16] = "\x39\x96\x2d\x55\xd1\x38\xef\xef\xc6\xe4\xa1\x4d\x5c\xd7\x35\x2e";

    for(i=0; i<16; i++) ver[i] ^= xor[15-i];
    decrypt( passwd, block_num, in, out );
    
    pp_array("out: ", out, 16);
    pp_array("ver: ", ver, 16);

    for(i=0; i<16; i++)
        assert( out[i] == ver[i] );
    for(i=0; i<16; i++)
        out[i] ^= xor[15-i];
    out[16] = 0;
    pp_array("out unxored: ", out, 16);
    printf("%s", out);
    return 0;
}

