import struct

extfunc = {
    0x12568: {'hint': 674, 'name': 'GetProcessHeap' },
    0x1257a: {'hint': 1362, 'name': 'Sleep' },
    0x12582: {'hint': 1434, 'name': 'VerifyVersionInfoW' },
    0x12598: {'hint': 669, 'name': 'GetProcAddress' },
    0x125aa: {'hint': 933, 'name': 'LoadLibraryA' },
    0x125ba: {'hint': 804, 'name': 'GlobalMemoryStatusEx' },
    0x125d2: {'hint': 720, 'name': 'GetSystemInfo' },
    0x125e2: {'hint': 1017, 'name': 'OutputDebugStringA' },
    0x125f8: {'hint': 538, 'name': 'GetDiskFreeSpaceExW' },
    0x1260e: {'hint': 186, 'name': 'CreateFileA' },
    0x1261c: {'hint': 819, 'name': 'HeapFree' },
    0x12b68: {'hint': 127, 'name': 'CloseHandle' },
    0x12b58: {'hint': 1504, 'name': 'WriteConsoleW' },
    0x12b44: {'hint': 1277, 'name': 'SetFilePointerEx' },
    0x12554: {'hint': 615, 'name': 'GetModuleHandleW' },
    0x12b22: {'hint': 494, 'name': 'GetConsoleMode' },
    0x12b12: {'hint': 476, 'name': 'GetConsoleCP' },
    0x12afe: {'hint': 402, 'name': 'FlushFileBuffers' },
    0x12aee: {'hint': 918, 'name': 'LCMapStringW' },
    0x12ae2: {'hint': 824, 'name': 'HeapSize' },
    0x12ad0: {'hint': 709, 'name': 'GetStringTypeW' },
    0x12ac2: {'hint': 822, 'name': 'HeapReAlloc' },
    0x12aac: {'hint': 1018, 'name': 'OutputDebugStringW' },
    0x12540: {'hint': 521, 'name': 'GetCurrentProcess' },
    0x1252a: {'hint': 1430, 'name': 'VerSetConditionMask' },
    0x1250e: {'hint': 1263, 'name': 'SetErrorMode' },
    0x1251e: {'hint': 815, 'name': 'HeapAlloc' },
    0x12b34: {'hint': 1314, 'name': 'SetStdHandle' },
    0x12500: {'hint': 337, 'name': 'ExitProcess' },
    0x127c0: {'hint': 726, 'name': 'GetSystemTimeAsFileTime' },
    0x127da: {'hint': 457, 'name': 'GetCommandLineW' },
    0x127ec: {'hint': 871, 'name': 'IsDebuggerPresent' },
    0x12800: {'hint': 877, 'name': 'IsProcessorFeaturePresent' },
    0x1281c: {'hint': 592, 'name': 'GetLastError' },
    0x1282c: {'hint': 1291, 'name': 'SetLastError' },
    0x1283c: {'hint': 526, 'name': 'GetCurrentThreadId' },
    0x12852: {'hint': 289, 'name': 'EncodePointer' },
    0x12862: {'hint': 254, 'name': 'DecodePointer' },
    0x12872: {'hint': 614, 'name': 'GetModuleHandleExW' },
    0x12888: {'hint': 977, 'name': 'MultiByteToWideChar' },
    0x1289e: {'hint': 1485, 'name': 'WideCharToMultiByte' },
    0x128b4: {'hint': 704, 'name': 'GetStdHandle' },
    0x128c4: {'hint': 574, 'name': 'GetFileType' },
    0x128d2: {'hint': 261, 'name': 'DeleteCriticalSection' },
    0x128ea: {'hint': 702, 'name': 'GetStartupInfoW' },
    0x128fc: {'hint': 611, 'name': 'GetModuleFileNameW' },
    0x12912: {'hint': 1505, 'name': 'WriteFile' },
    0x1291e: {'hint': 1069, 'name': 'QueryPerformanceCounter' },
    0x12938: {'hint': 522, 'name': 'GetCurrentProcessId' },
    0x1294e: {'hint': 551, 'name': 'GetEnvironmentStringsW' },
    0x12968: {'hint': 413, 'name': 'FreeEnvironmentStringsW' },
    0x12982: {'hint': 1410, 'name': 'UnhandledExceptionFilter' },
    0x1299e: {'hint': 1347, 'name': 'SetUnhandledExceptionFilter' },
    0x129bc: {'hint': 840, 'name': 'InitializeCriticalSectionAndSpinCount' },
    0x129e4: {'hint': 1377, 'name': 'TerminateProcess' },
    0x129f8: {'hint': 1395, 'name': 'TlsAlloc' },
    0x12a04: {'hint': 1397, 'name': 'TlsGetValue' },
    0x12a12: {'hint': 1398, 'name': 'TlsSetValue' },
    0x12a20: {'hint': 1396, 'name': 'TlsFree' },
    0x12a2a: {'hint': 293, 'name': 'EnterCriticalSection' },
    0x12a42: {'hint': 930, 'name': 'LeaveCriticalSection' },
    0x12a5a: {'hint': 882, 'name': 'IsValidCodePage' },
    0x12a6c: {'hint': 420, 'name': 'GetACP' },
    0x12a76: {'hint': 646, 'name': 'GetOEMCP' },
    0x12a82: {'hint': 435, 'name': 'GetCPInfo' },
    0x12a8e: {'hint': 935, 'name': 'LoadLibraryExW' },
    0x12aa0: {'hint': 1197, 'name': 'RtlUnwind' },
    0x12b76: {'hint': 194, 'name': 'CreateFileW' },
    0x1275a: {'hint': 173, 'name': 'DestroyWindow' },
    0x1274c: {'hint': 371, 'name': 'GetMessageW' },
    0x1273a: {'hint': 625, 'name': 'PostQuitMessage' },
    0x12726: {'hint': 831, 'name': 'TranslateMessage' },
    0x12712: {'hint': 510, 'name': 'IsDialogMessageW' },
    0x12704: {'hint': 582, 'name': 'MessageBoxA' },
    0x126f6: {'hint': 316, 'name': 'GetDlgItem' },
    0x126e0: {'hint': 691, 'name': 'SendDlgItemMessageW' },
    0x126ce: {'hint': 62, 'name': 'CheckDlgButton' },
    0x126c0: {'hint': 800, 'name': 'ShowWindow' },
    0x126b2: {'hint': 541, 'name': 'LoadBitmapW' },
    0x1269c: {'hint': 102, 'name': 'CreateDialogParamW' },
    0x1268e: {'hint': 589, 'name': 'MessageBoxW' },
    0x1267c: {'hint': 787, 'name': 'SetWindowTextA' },
    0x1266c: {'hint': 700, 'name': 'SendMessageW' },
    0x1265c: {'hint': 228, 'name': 'EnableWindow' },
    0x1264a: {'hint': 318, 'name': 'GetDlgItemTextA' },
    0x12636: {'hint': 181, 'name': 'DispatchMessageW' },
    0x1278c: {'hint': 560, 'name': 'GetPixel' },
    0x12776: {'hint': 49, 'name': 'CreateCompatibleDC' },
    0x12798: {'hint': 725, 'name': 'SelectObject' },
    0x80000011: {'hint': 17, 'name': '<none>' },
}

anno_addr = {
    0x4012f0: "# function to reset the checkboxes state",
    0x4011c0: "# set the dialog main button text",
    0x401522: "#here arrives the checkbox clicks and button press",
    0x401c70: "# function that returns 0xf4b00f4b in eax",
    0x401ca0: "# function that returns 0xf4b01f4b in eax",
    0x401cd0: "# function that returns 0xf4b02f4b in eax",
    0x401d00: "# function that returns 0xf4b03f4b in eax",
    0x4165e8: "# Address of buffer for the input of the passwd (33bytes)",

    0x402180: "# linklist_search 1()",
    0x402210: "# linklist_search 2()",
    0x416628: "# 32bytes allocated with HeapAlloc (malloc)",
    0x416638: "# link list root",
}

anno_hex_value = {
}

data = open("unstable.machines.exe", "rb").read()
disasm_lines = open("unstable.machines.exe.disasm", "rb").readlines()

rodata = data #[0xc600:0xc600+0x4b84]

assert( 0x126f6 in extfunc)
end = 0xc600+0x4b84

b = 0xc600 #0xd308
while b < end:
    
    while rodata[b] == '\0':
        b += 1

    f = b + 1
    while rodata[f] != '\0':
        f += 1
    rostr = str(rodata[b:f])
    addr = b - 0xc600 + 0x40e000

    straddr = "0x%x" % addr
    cnt = 0
    for i, l in enumerate(disasm_lines):   
        if straddr in l:
            #if 'call   *0x' in l:
            #    fid, = struct.unpack("<I", rodata[b:b+4])
            #    if fid in extfunc:
            #        rostr = extfunc[fid]['name']
            #    else:
            #        print(fid," not found")
            newl = l.rstrip() + "\t"*3 + "# cst str @0x%x: '%s'\n" % (b, rostr)
            disasm_lines[i] = newl
            cnt += 1
    print("found string '%s': replaced %d times\n" % (rostr, cnt))
    b = f

b = 0xc600 #0xd308
while b < end:
    
    addr = b - 0xc600 + 0x40e000
    fid, = struct.unpack("<I", rodata[b:b+4])
    if fid in extfunc:
        rostr = extfunc[fid]['name']

        straddr = "0x%x" % addr
        cnt = 0
        for i, l in enumerate(disasm_lines):   
            if straddr in l:
                newl = l.rstrip() + "\t"*3 + "# cst str @0x%x: '%s'\n" % (b, rostr)
                disasm_lines[i] = newl
                cnt += 1
        print("found string '%s': replaced %d times\n" % (rostr, cnt))

    b += 4


out = open("unstable.machines.exe.disasm2", "wt")
for l in disasm_lines:
    out.write( l )
