// compile: gcc -m64 -march=native -funroll-loops -O2 waterscript.c -o waterscript

#include <inttypes.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>


typedef uint32_t u32;
typedef uint64_t u64;
typedef uint8_t  u8;

const char *hexchr = "0123456789abcdef";

// special mem that seems const
static const u32 mem_0xf4b = 0xd2d413b1;

u32 target_values_unscrambles[8];

const u32 mem_cst[] = {
    0x9aacc69b,
    0x89a31d8a,
    0xa5b2cd1c,
    //0xa5b3cd5c,
    0xb0a5ce54,
};

// This memory represent part of the linklist last node buffer, 
// from 0x1000 (0x139908 + 0x1000)
static u8  memory[0x1000];
static inline u32* mem(int idx) {
    //printf("%d\n", idx);
    //assert(idx < sizeof(memory));
    return (u32*)(memory + idx);
}

static inline u32 scramble( u32 v, u32 sh ) {
    u32 v1 = v << (sh&0xff);
    u32 v2 = v >> ((sh&0xff)+1);
    //return (v1 ^ v2)
    return (v1 ^ v2) & 0xffffffff;
}

// Clear bits of v that are not set in n
static u32 clr( u32 v, u32 n ) {
    u32 nv = ~v;
    nv |= ~n;
    return ~nv;
}

static void dumpmem(char *s, int off, int len) {
    int i;
    puts(s);
    for(i=off; i<off+len; i+=4) {
        u8 *p = (u8*)mem(i);
        printf("0x%02x:  0x%08x %c%c%c%c\n", i, *mem(i), p[0], p[1], p[2], p[3]);
    }
}

static int scramble_passwd(const int print) {
    //const int print=0;
    int i;

    //assert( *mem(0xf4b) == mem_0xf4b );

    // Initialize the mem with hash values that are pre-computed 
    // (I simply got them using gdb at the right time and inspecting memory)
    *mem(0x0) = mem_cst[0];
    *mem(0x4) = mem_cst[1];
    *mem(0x8) = mem_cst[2];
    *mem(0xc) = mem_cst[3];

    memset( memory+0x40, 0, 0x20 );

    if( print )
        dumpmem("Passwd original:", 0x20, 32);

    *mem(0x5c) = 0;
    while( *mem(0x5c) != 4)
    {
        *mem(0x44) = *mem(0x5c) * 2;
        *mem(0x4c) = *mem(0x20 + *mem(0x44)*4);

        *mem(0x44) += 1;
        *mem(0x50) = *mem(0x20 + *mem(0x44)*4);

        *mem(0x54) = 0;
        *mem(0x58) = 0;
        while( *mem(0x58) != 64 ) {
            *mem(0x40) = scramble( *mem(0x50), 4 );
            *mem(0x60) = *mem(0x40) + *mem(0x50);

            // (pix_y  60)

            // *mem(0x44) = *mem(0x54)
            // *mem(0x48) = 0
            *mem(0x64) = *mem(0x00 + clr(*mem(0x54)>>0, 3)*4);
            *mem(0x64) += *mem(0x54);
            *mem(0x60) ^= *mem(0x64);
            *mem(0x4c) += *mem(0x60);
            // mem_4c += (mem_50 + scramble(mem_50)) ^ (mem_54 + mem_[mem_54 & 3])

            // (pix_y  81)

            *mem(0x40) = mem_0xf4b + 2;
            *mem(0x54) += *mem(0x40);

            // (pix_y  87)

            // *mem(0x44) = *mem(0x4c)
            // *mem(0x48) = 4
            *mem(0x40) = scramble( *mem(0x4c), 4 );
            *mem(0x60) = *mem(0x40) + *mem(0x4c);

            // *mem(0x44) = *mem(0x54)
            // *mem(0x48) = 11
            *mem(0x64) = *mem(0x00 + clr(*mem(0x54)>>11, 3)*4);
            *mem(0x64) += *mem(0x54);
            *mem(0x60) ^= *mem(0x64);
            *mem(0x50) += *mem(0x60);
            // mem_50 += (mem_4c + scramble(mem_4c)) ^ (mem_54 + mem_[mem_54>>11 & 3])

            // (pix_y 123)

            *mem(0x58) = *mem(0x58) + 1;
        }

        *mem(0x20 + (*mem(0x5c))*2*4) = *mem(0x4c);
        *mem(0x20 + (*mem(0x5c)*2+1)*4) = *mem(0x50);
    
        // (pix_y 156)

        *mem(0x5c) = *mem(0x5c) + 1;
    }

    // (pix_y 165)
    // sleep + return 

    if( print )
        dumpmem("Passwd after scrambling1:", 0x20, 32);

    u64 rdx = 0x756af83de1ffe263;
    u64 rcx = 0xfa34ae1002547b89;
    u64 r8;
    u64 *p = (u64*)mem(0x20);

    rdx = ~rdx;
    #define BYTE(n) ((rcx)>>(8*(n)) & 0xff)
    rcx =   (BYTE(0)<<56) | (BYTE(1)<<48) | (BYTE(2)<<40) | (BYTE(3)<<32) |
            (BYTE(4)<<24) | (BYTE(5)<<16) | (BYTE(6)<<8)  | (BYTE(7)<<0);
    #undef BYTE
    r8 = *p;
    *p = (r8 ^ rcx) + rdx;
    p++;

    #define BYTE(n) ((r8)>>(8*(n)) & 0xff)
    r8 = *p;
    r8 ^= rdx;
    r8 =    (BYTE(0)<<56) | (BYTE(1)<<48) |(BYTE(2)<<40) |(BYTE(3)<<32) |
            (BYTE(4)<<24) | (BYTE(5)<<16) |(BYTE(6)<<8)  |(BYTE(7)<<0);
    r8 -= rcx;
    *p = r8;
    p++;
    #undef BYTE

    #define BYTE(n) ((r8)>>(8*(n)) & 0xff)
    #define ROR(v,n) (((v)>>(n)) | ((v)<<(64-(n))))
    #define ROL(v,n) (((v)<<(n)) | ((v)>>(64-(n))))
    r8 = *p;
    r8 =    (BYTE(0)<<56) | (BYTE(1)<<48) |(BYTE(2)<<40) |(BYTE(3)<<32) |
            (BYTE(4)<<24) | (BYTE(5)<<16) |(BYTE(6)<<8)  |(BYTE(7)<<0);
    rcx = ROL(rcx, 0xd);
    r8 = ROR(r8, rcx&0xff);
    *p = r8;
    p++;
    #undef BYTE

    r8 = *p;
    r8 ^= rcx;
    rdx = ROR(rdx, 0x15);
    u64 tmp = rdx; rdx = rcx; rcx = tmp;
    r8 = ROR(r8, rcx&0xff);
    r8 += rdx;
    *p = r8;
    p++;

    if( print )
        dumpmem("Passwd after scrambling2:", 0x20, 32);

    return 0;
}

static int unscramble_passwd(const int print) {
    //const int print=0;
    int i;

    *mem(0x0) = mem_cst[0];
    *mem(0x4) = mem_cst[1];
    *mem(0x8) = mem_cst[2];
    *mem(0xc) = mem_cst[3];

    if( print )
        dumpmem("Passwd before unscrambling:", 0x20, 32);

    u64 rdx = 0x756af83de1ffe263;
    u64 rcx = 0xfa34ae1002547b89;
    u64 r8;
    const u64 *p = (u64*)mem(0x20);
    u64 *p_out = (u64*)mem(0x20);

    rdx = ~rdx;
    #define BYTE(n) ((rcx)>>(8*(n)) & 0xff)
    rcx =   (BYTE(0)<<56) | (BYTE(1)<<48) | (BYTE(2)<<40) | (BYTE(3)<<32) |
            (BYTE(4)<<24) | (BYTE(5)<<16) | (BYTE(6)<<8)  | (BYTE(7)<<0);
    #undef BYTE
    r8 = *p;
    *p_out = (r8 - rdx) ^ rcx;
    //*p = (r8 ^ rcx) + rdx;
    p++;
    p_out++;

    #define BYTE(n) ((r8)>>(8*(n)) & 0xff)
    r8 = *p;
/*
    r8 ^= rdx;
    r8 =    (BYTE(0)<<56) | (BYTE(1)<<48) |(BYTE(2)<<40) |(BYTE(3)<<32) |
            (BYTE(4)<<24) | (BYTE(5)<<16) |(BYTE(6)<<8)  |(BYTE(7)<<0);
    r8 -= rcx;
*/
    r8 += rcx;
    r8 =    (BYTE(0)<<56) | (BYTE(1)<<48) |(BYTE(2)<<40) |(BYTE(3)<<32) |
            (BYTE(4)<<24) | (BYTE(5)<<16) |(BYTE(6)<<8)  |(BYTE(7)<<0);
    r8 ^= rdx;
    *p_out = r8;
    p++;
    p_out++;
    #undef BYTE

    #define BYTE(n) ((r8)>>(8*(n)) & 0xff)
    #define ROR(v,n) (((v)>>(n)) | ((v)<<(64-(n))))
    #define ROL(v,n) (((v)<<(n)) | ((v)>>(64-(n))))
    r8 = *p;
    rcx = ROL(rcx, 0xd);
    r8 = ROL(r8, rcx&0xff);
    r8 =    (BYTE(0)<<56) | (BYTE(1)<<48) |(BYTE(2)<<40) |(BYTE(3)<<32) |
            (BYTE(4)<<24) | (BYTE(5)<<16) |(BYTE(6)<<8)  |(BYTE(7)<<0);
    *p_out = r8;
    p++;
    p_out++;
    #undef BYTE

    r8 = *p;
    rdx = ROR(rdx, 0x15);
    u64 tmp = rdx; rdx = rcx; rcx = tmp;
    r8 -= rdx;
    r8 = ROL(r8, rcx&0xff);

    r8 ^= rdx; // rdx is the old rcx
    *p_out = r8;
/*
    r8 ^= rcx;
    rdx = ROR(rdx, 0x15);
    u64 tmp = rdx; rdx = rcx; rcx = tmp;
    r8 = ROR(r8, rcx&0xff);
    r8 += rdx;
*/

    if( print )
        dumpmem("Passwd after unscrambling2:", 0x20, 32);


    *mem(0x5c) = 0;
    while( *mem(0x5c) != 4)
    {
        //*mem(0x5c) = block;
        *mem(0x54) = (mem_0xf4b + 2)*64;

        *mem(0x4c) = *mem(0x20 + (*mem(0x5c))*2*4);
        *mem(0x50) = *mem(0x20 + (*mem(0x5c)*2+1)*4);

        for(i=0; i<64; i++) {

            *mem(0x40) = scramble( *mem(0x4c), 4 );
            *mem(0x60) = *mem(0x40) + *mem(0x4c);

            // *mem(0x44) = *mem(0x54)
            // *mem(0x48) = 11
            *mem(0x64) = *mem(0x00 + clr(*mem(0x54)>>11, 3)*4);
            *mem(0x64) += *mem(0x54);
            *mem(0x60) ^= *mem(0x64);
            *mem(0x50) -= *mem(0x60);

            *mem(0x40) = mem_0xf4b + 2;
            *mem(0x54) -= *mem(0x40);


            *mem(0x40) = scramble( *mem(0x50), 4 );
            *mem(0x60) = *mem(0x40) + *mem(0x50);

            // (pix_y  60)

            // *mem(0x44) = *mem(0x54)
            // *mem(0x48) = 0
            *mem(0x64) = *mem(0x00 + clr(*mem(0x54)>>0, 3)*4);
            *mem(0x64) += *mem(0x54);
            *mem(0x60) ^= *mem(0x64);
            *mem(0x4c) -= *mem(0x60);
        }

        assert( *mem(0x54) == 0 );

        *mem(0x20 + (*mem(0x5c))*2*4) = *mem(0x4c);
        *mem(0x20 + (*mem(0x5c)*2+1)*4) = *mem(0x50);

        *mem(0x5c) = *mem(0x5c) + 1;
    }

    if( print )
        dumpmem("Passwd after unscrambling1:", 0x20, 32);
}

static void find_passwd()
{
    u32 mem_54 = 0;
    u32 tgt_mem_4c = 0x771069e2;
    u32 tgt_mem_50 = 0x82146b05;

    //u32 mem_4c = 0x444c4f46;
    //u32 mem_50 = 0x54534946;

    u64 i;
    for(i=0x0; i<0x100000000ULL; i++) {
        int j;
            /*
        u64 v = 0;
        for(j=0; j<8; j++) {
            v <<= 8;
            v |= hexchr[ (i>>(4*j)) & 0xf ];
        }

        u32 mem_4c = v & 0xffffffff;
        u32 mem_50 = v >> 32;
        */
        u32 mem_4c = i;
        u32 mem_50 = 0x54534946;

        u32 mem_54 = 0;
        for(j=0; j<64; j++) {
            mem_4c += (mem_50 + scramble(mem_50, 4)) ^ (mem_54 + mem_cst[mem_54 & 3]);
                                                  // ^ 0x9aacc69b
            mem_54 += mem_0xf4b + 2;
            mem_50 += (mem_4c + scramble(mem_4c, 4)) ^ (mem_54 + mem_cst[(mem_54>>11) & 3]);
                                                 // ^ 0xd2d413b1 + 2 + 0xa5b3cd5c
                                                 // ^ 0x7887e10f
        }

        if( mem_4c == tgt_mem_4c && mem_50 == tgt_mem_50 ) {
            printf("bingo! 0x%llx\n", i);
        }
        if( (i & 0xffffff) == 0)
            printf(". i=0x%lx\n", i);
    }

    //printf("0x%08x 0x%08x\n", mem_4c, mem_50);
}


static void find_scramble()
{
    u32 mem_54 = 0;
    u32 orig_mem_4c = 0x34c0425e; //0x82c9c750;
    u32 orig_mem_50 = 0x680ff193; //0xbd060a1f;

    mem_54 = 0;
    u32 next_mem_4c = orig_mem_4c + ((orig_mem_50 + scramble(orig_mem_50, 4)) ^ (mem_54 + mem_cst[mem_54 & 3]));
    mem_54 += mem_0xf4b + 2;
    u32 next_mem_50 = orig_mem_50 + ((next_mem_4c + scramble(next_mem_4c, 4)) ^ (mem_54 + mem_cst[(mem_54>>11) & 3]));

    u64 i;
    for(i=0; i<0x100000000ULL; i++) {
        mem_54 = 0;
        u32 mem_50 = i;
        u32 new_orig_mem_4c = next_mem_4c - ((mem_50 + scramble(mem_50, 4)) ^ (mem_54 + mem_cst[mem_54 & 3]));
        u32 mem_4c = new_orig_mem_4c + ((mem_50 + scramble(mem_50, 4)) ^ (mem_54 + mem_cst[mem_54 & 3]));
        assert( mem_4c == next_mem_4c );
        mem_54 += mem_0xf4b + 2;
        u32 new_mem_50 = mem_50 + ((mem_4c + scramble(mem_4c, 4)) ^ (mem_54 + mem_cst[(mem_54>>11) & 3]));

        if( mem_4c == next_mem_4c && new_mem_50 == next_mem_50 ) {
            printf("bingo! 0x%llx\n", i);
        }
        if( (i & 0xffffff) == 0)
            printf(". i=0x%lx\n", i);
    }
}

void test() {
    const int MAX = 0x10000;
    int i, j;
    //u8 passwd[32];
    u8* passwd = malloc( MAX * 32 );
    for(i=0; i<MAX; i++) {
        for(j=0; j<32; j++)
            passwd[(i*32)+j] = rand(); //(j * 0xc89ef539) % 257;
    }

    u8* scra = malloc( MAX * 32 );
    for(i=0; i<MAX; i++) {
        memcpy( memory+0x20, passwd+(i*32), 32 );
        scramble_passwd(0);
        memcpy(scra+(i*32), memory+0x20, 32 );
    }

    for(i=0; i<MAX; i++) {
        memcpy( memory+0x20, scra+(i*32), 32 );
        unscramble_passwd(0);

        for(j=0; j<32; j++)
            assert( *(u8*)mem( 0x20 + j ) == passwd[(i*32)+j] );
    }
}

int main(int argc, char *argv[]) {
    const int tot = 0x68; //0x100; //0x1000;
    char memory_orig[tot];

    int fd = open("nodeLast.bin", O_RDONLY);

    int off = 0;
    lseek( fd, 0x1000, SEEK_SET );
    while( off < tot ) {
        int r = read( fd, memory_orig+off, tot-off );
        if( r > 0 )
            off += r;
    }

    srand( time( NULL ) );
    

    //find_scramble();
    //find_passwd();
    //test();
    //return 0;

    //unscramble_target();

    //char *passwd = "00112233445566778899AABBCCDDEEFF";
    char passwd[32];
    int block = 0;

    u32 GOLDFIST_values[] = {
        0xf13026b7, 0x29e3d029, 0x1811c70b, 0x7cc1161c,
        0xd31f3442, 0xd401d570, 0x37553718, 0x6580090b,
    };

    u32 found_passwd[] = {
        0x82c9c750, 0xbd060a1f, 0x34c0425e, 0x680ff193,
        0xe443ff97, 0xafbf184a, 0x5b49bb67, 0xd5bc253f,
    }; 
    const u32 target_values[] = {
        0x29816725,  0xd4e46b75,  0x349f7ce1,  0x331de826,
        0xc4e3648e,  0x7a658d66,  0xe763e820,  0xe6dd1abf,
    };

    // values for 00112233445566778899AABBCCDDEEFF
    u32 some_values[] = {
        0xd9d0a94a,  0x96aff9d1,  0x1715b48c,  0xbad51448,
        0x5a5c8a6a,  0x8e949018,  0x6ad7b645,  0x7ee5afef,
    };


    //found_passwd[0] = 0xafe0c38e;
    //found_passwd[1] = 0x723c3be0;
    
    memcpy(memory, memory_orig, tot);
    int j;
    //int i = 0x196a4a75, j;
    //for(j=0; j<8; j++)
    //    passwd[(block*8)+j] = hexchr[ (i>>(4*j)) & 0xf ];
    //for(j=0; j<32; j++)
    //    passwd[j] = (j * 0xc89ef539) % 257;

    assert( *mem(0x20) == 0x444c4f47 );
    //memcpy( memory+0x20, passwd, 32 );
    //memcpy( memory+0x20, "FOLDFIST                        ", 32 );
    //memcpy( memory+0x20, "GOLDFIST", 8 );
    //memcpy( memory+0x20, "363c 1632 3616 3237 1733 3716 3236 1733", 32 );
    //memcpy( memory+0x20, "00112233445566778899AABBCCDDEEFF", 32 );
    //memcpy( memory+0x20, found_passwd, 32 );
    //scramble_passwd(1);
    puts("");
    memcpy( memory+0x20, target_values, 32);
    //memcpy( memory+0x20, some_values, 32);
    //memcpy( memory+0x20, space32_values, 32);
    //memcpy( memory+0x20, GOLDFIST_values, 32);
    unscramble_passwd(1);

    //puts("passwd in ascii:");
    return 0;
}

